class Error(Exception):
    """Base class for other exceptions"""
    pass

class SetuptoolFailure(Error):
    """Raised when setuptool encounters an error"""
    def __init__(self, message, errors='002'):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.errors = errors

class PackageJsonNotFound(Error):
    """Raised when package.json is not found"""
    def __init__(self, message, errors='001'):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.errors = errors

class NoSetupConfiguration(Error):
    """Raised when 'setup' key in package.json is not found"""
    def __init__(self, message, errors='003'):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.errors = errors



import json, shlex, subprocess, io, sys, os
import os.path
from .post_operations import update_package_json_after_uninstall, update_package_json_after_operation
from .setup import Setup

cur_path = str(os.getcwd())
setuptool = Setup()

class PyPM:
    def __init__(self):
        self.package_json = None

    def set_variables(self, path=cur_path, verbose=False, service='pip', arguments=None):
        def args():
            return ' '.join(arguments.split(','))
        self.path = path
        self.verbose = verbose
        self.service = service
        self.arguments = [args() if arguments != None and ',' in arguments else arguments][0]

    def __reader__(self):
        try:
            with open(os.path.join(self.path, 'package.json'), 'r', encoding='utf8') as packagejson:
                return json.loads(packagejson.read())
        except FileNotFoundError:
            print('No package.json found. Please run pypm init')
            exit()

    def __commander__(self, key, item):
        arguments = self.arguments if self.arguments != None else ''
        option = {
            'run': shlex.split(f'{item} {arguments}'),
            'install': shlex.split(f'{self.service} install {arguments} {item}'),
            'uninstall': shlex.split(f'{self.service} uninstall {arguments} {item}'),
            'update': shlex.split(f'{self.service} install --upgrade {item}') if 'pip' in self.service else shlex.split(f'{self.service} update {arguments} {item}')
        }.get(key)
        return option

    def __get_dependency_version__(self, temp_dependency_list, dependency_list):
        final_dependency_list = list()
        for key in temp_dependency_list:
            final_dependency_list.append(f'{key}=={self.package_json[dependency_list][key]}')
        return final_dependency_list
    
    def __get_dependencies__(self, update=False):
        dependencies = list()
        hasErrors = False
        try:
            if update:
                dependencies = list(self.package_json['dependencies'].keys()).extend(list(self.package_json['devDependencies']).keys())
            else:
                temp_list_one = list(self.package_json['dependencies'].keys())
                dependencies = self.__get_dependency_version__(temp_list_one, 'dependencies')
                temp_list_two = list(self.package_json['devDependencies'].keys())
                dependencies.extend(self.__get_dependency_version__(temp_list_two, 'devDependencies'))
        except KeyError as ke:
            if ke == 'devDependencies':
                if update:
                    dependencies = list(self.package_json['dependencies'].keys())
                else:
                    temp_list_three = list(self.package_json['dependencies'].keys())
                    dependencies = self.__get_dependency_version__(temp_list_three, 'dependencies')
            else:
                hasErrors = True
        finally:
            if not hasErrors:
                return self.__list_to_str__(dependencies)
            print('No dependencies found.')
            sys.exit() 

    def __process__(self, cmd):
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        for line in io.TextIOWrapper(proc.stdout):
            print(line.replace('\n', ''))

    def __list_to_str__(self, item_list):
        return ' '.join(item_list)

    def __pretty_key__(self, key):
        return key.replace(' ', ', ')
    
    def __assign_package_json__(self):
        self.package_json = self.__reader__()

    def start(self):
        self.__assign_package_json__()
        if self.verbose:
            print('Running start command...\n')
        cmd = self.__commander__('run', str(self.package_json['scripts']['start']))
        self.__process__(cmd)
    
    def run(self, key):
        self.__assign_package_json__()
        if self.verbose:
            print(f'Running {key}...\n')
        cmd = self.__commander__('run', str(self.package_json['scripts'][str(key)]))
        self.__process__(cmd)
    
    def install(self, key):
        key = self.__list_to_str__(key)
        self.__assign_package_json__()
        if len(key) == 0:
            if self.verbose:
                print('Installilng all dependencies...')
            cmd = self.__commander__('install', self.__get_dependencies__())
        else:
            if self.verbose:
                print(f'Installing package(s) {self.__pretty_key__(key)}...\n')
            cmd = self.__commander__('install', key)
        self.__process__(cmd)
        if len(key) != 0:
            update_package_json_after_operation(key, self.path, self.package_json)
    
    def uninstall(self, key):
        key = self.__list_to_str__(key)
        self.__assign_package_json__()
        if len(key) == 0:
            if self.verbose:
                print('Uninstalling all dependencies...\n')
            cmd = self.__commander__('uninstall', self.__get_dependencies__())
        else:
            if self.verbose:
                print(f'Uninstalling package(s) {self.__pretty_key__(key)}...\n')
            cmd = self.__commander__('uninstall', key)
        self.__process__(cmd)
        if len(key) != 0:
            update_package_json_after_uninstall(key, self.path, self.package_json)
    
    def update(self, key):
        key = self.__list_to_str__(key)
        self.__assign_package_json__()
        if len(key) == 0:
            if self.verbose:
                print('Updating all dependencies...\n')
            cmd = self.__commander__('update', self.__get_dependencies__(update=True))
        else:
            if self.verbose:
                print(f'Updating package(s) {self.__pretty_key__(key)}...\n')
            cmd = self.__commander__('update', key)
        self.__process__(cmd)
        if len(key) != 0:
            update_package_json_after_operation(key, self.path, self.package_json)
    
    def setup_py(self, key):
        self.__assign_package_json__()
        setuptool.set_vars(pkg_json=self.package_json, update_packages=key)
        setuptool.configure()
        setuptool.begin()



import os.path as path
import os, json, sys, time
from .errors import SetuptoolFailure, NoSetupConfiguration
import io, shlex, subprocess, configparser, pkg_resources


class SetupGenerator:
    def __init__(self):
        self.config = configparser.ConfigParser()
    
    def prereqs(self):
        cmd = shlex.split('python3 -m pip install --upgrade pip setuptools wheel')
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        for line in io.TextIOWrapper(proc.stdout):
            print(line.replace('\n', ''))
        
    def set_generator_var(self, pkg_json):
        self.pkg_json = pkg_json
    
    def _get_package_resource(self, resource):
        resource_package = __name__
        resource_path = '/'.join(('data', resource))
        return pkg_resources.resource_string(resource_package, resource_path).decode('utf-8')

    def read(self):
        cfg = self._get_package_resource('setup.cfg')
        self.config.read_string(cfg)

    def write(self):
        setup_py = self._get_package_resource('setup.py')
        with open('setup.py', 'w') as setup_py_file:
            setup_py_file.write(setup_py)

        with open('setup.cfg', 'w') as configfile:
            self.config.write(configfile)

    def _custom_key_(self, root_key, vkey):
        package_data_string = ''
        if vkey == 'package_data':
            for key, value in self.pkg_json[vkey].items():
                list_to_str = '\n'.join(value)
                package_data_string += f"{key} = {list_to_str}"
        else:
            self.config[root_key] = self.pkg_json[vkey]
    
    def _custom_key_join(self, root_key, vkey):
        self.config[root_key][vkey] = '\n'.join(self.pkg_json[vkey])
    
    def generate(self):
        content = {
            'metadata': ['name', 'author', 'author_email', 'version', 'description', 'long_description', 'url', 'keywords',
                'license', 'classifiers', 'long_description_content_type'],
            'options': ['zip_safe', 'include_package_data', 'package_dir',
                'packages', 'scripts', 'install_requires', 'entry_points'],
            'options.package_data': ['package_data'],
            'options.extras_require': [],
            'options.packages.find': ['packages'],
            'options.data_files': ['data_files']
        }
        for key, value_keys in content.items():
            for vkey in value_keys:
                try:
                    if vkey == 'classifiers' or vkey == 'data_files' or vkey == 'keywords' or vkey == 'install_requires':
                        self._custom_key_join(key, vkey)
                    elif vkey == 'package_data':
                        self._custom_key_(key, vkey)
                    elif vkey == 'packages' and self.pkg_json['packages'] == []:
                        # if packages aren't specified, let setuptools find automatically
                        pass
                    elif key == 'options.packages.find' and self.pkg_json['package_dir']:
                        self.config[key]['where'] = self.pkg_json['package_dir'].replace('=', '')
                    else:
                        self.config[key][vkey] = self.pkg_json[vkey]
                except KeyError as ke:
                    if 'scripts' in str(ke):
                        del self.config['options']['scripts']
                    
                    if 'package_dir' in str(key):
                        del self.config['options']['package_dir']
                except TypeError as te:
                    pass

    
    def run_setup(self):
        cmd = shlex.split('pip install --editable .')
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        for line in io.TextIOWrapper(proc.stdout):
            print(line.replace('\n', ''))


class Setup:
    def __init__(self):
        self.pkg_json = None
        self.generator = SetupGenerator()

    def configure(self):
        try:
            self.pkg_json['setup']['name'] = self.pkg_json['name']
            self.pkg_json['setup']['license']  = self.pkg_json['license']
            self.pkg_json['setup']['version'] = self.pkg_json['version']
            self.pkg_json['setup']['author'] = self.pkg_json['author']
            self.pkg_json['setup']['description'] = self.pkg_json['description']
        except KeyError as ke:
            raise NoSetupConfiguration(ke)
            sys.exit()
        
        try:
            install_requires = []
            for dependency, version in self.pkg_json['dependencies'].items():
                install_requires.append(f'{dependency}=={version}')
            self.pkg_json['setup']['install_requires'] = install_requires
        except KeyError as ke:
            pass

    def set_vars(self, pkg_json, update_packages):
        self.pkg_json = pkg_json
        self.update_packages = update_packages
    
    def begin(self):
        try:
            self.generator.set_generator_var(self.pkg_json['setup'])
            if self.update_packages:
                self.generator.prereqs()
            self.generator.read()
            self.generator.generate()
            self.generator.write()
            print('Awaiting setup...')
            time.sleep(3)
            self.generator.run_setup()
            print('Setup complete.')
        except Exception as e:
            raise SetuptoolFailure(str(e))



import os, io, subprocess, shlex, re
from .generator import Generator

generator = Generator()

def _update_package_json(pkgjson, path):
    generator.set_variables(path, False)
    generator.__writer__(pkgjson)


def update_package_json_after_operation(dependency, path, pkgjson):
    obj =  {}
    count = 2
    errors = []
    command = shlex.split(f'pip freeze')
    proc = subprocess.Popen(command, stdout=subprocess.PIPE, env=os.environ.copy(), stderr=subprocess.PIPE)
    for line in io.TextIOWrapper(proc.stdout, encoding='utf8', newline=''):
        if dependency in line:
            en_v = re.sub("['\"]", '', line.replace('\n', ''))
            if en_v.find('>') == -1:
                idx = en_v.find('=')
            else:
                idx = en_v.find('>')
            obj[re.sub('\s+', '', str(en_v[:idx]))] = str(en_v[idx+count:])
    try:
        pkgjson['dependencies'].update(obj)
    except KeyError:
        pkgjson['devDependencies'].update(obj)
    except TypeError as te:
        print(f'Error occurred updating package.json. {dependency} was not found.')
        errors.append(te)
    finally:
        if len(errors) == 0:
            _update_package_json(pkgjson, path)


def update_package_json_after_uninstall(dependency, path, pkgjson):
    errors = []
    try:
        pkgjson['dependencies'][str(dependency)]
        del pkgjson['dependencies'][str(dependency)]
    except KeyError:
        del pkgjson['devDependencies'][str(dependency)]
    except TypeError as te:
        print(f'Error occurred updating package.json. {dependency} was not found.')
        errors.append(te)
    finally:
        if len(errors) == 0:
            _update_package_json(pkgjson, path)
        

